@@ -5,7 +5,7 @@
 
 package XML::Compile::Schema::Template;
 use vars '$VERSION';
-$VERSION = '0.87';
+$VERSION = '0.88';
 
 
 use XML::Compile::Schema::XmlWriter;
@@ -67,7 +67,8 @@ sub _block($@)
     sub { my @elems  = map { $_->() } odd_elements @pairs;
           my @tags   = map { $_->{tag} } @elems;
 
-          my $struct = "$block of ". join(', ',@tags);
+          local $" = ', ';
+          my $struct = @tags ? "$block of @tags" : "empty $block";
           my @lines;
           while(length $struct > 65)
           {   $struct =~ s/(.{1,60})(\s)//;
@@ -76,7 +77,6 @@ sub _block($@)
           push @lines, $struct;
           $lines[$_] =~ s/^/  / for 1..$#lines;
 
-          local $" = ', ';
            { tag    => $block
            , elems  => \@elems
            , struct => \@lines
@@ -115,7 +115,7 @@ sub block_handler
 
 sub element_handler
 {   my ($path, $args, $label, $min, $max, $req, $opt) = @_;
-    sub { my $data = $opt->();
+    sub { my $data = $opt->() or return;
           my $occur
            = $max eq 'unbounded' && $min==0 ? 'occurs any number of times'
            : $max ne 'unbounded' && $max==1 && $min==0 ? 'is optional' 
@@ -158,6 +158,11 @@ sub element_nillable
     sub { (occur => "$childname is nillable", $do->()) };
 }
 
+sub element_abstract
+{   my ($path, $args, $ns, $childname, $do) = @_;
+    sub { () };
+}
+
 my %recurse;
 sub complex_element
 {   my ($path, $args, $tag, $elems, $attrs, $any_attr) = @_;
@@ -345,13 +350,13 @@ sub attribute_fixed
 }
 
 sub substgroup
-{   my ($path, $args, $type, %do) = @_;
-    my ($first, $do) = keys %do;
+{   my ($path, $args, $type, @do) = @_;
+    my @tags = sort map { $_->[0] } odd_elements @do;
+
     sub { +{ kind    => 'substitution group'
-           , tag     => (unpack_type $type)[1]
-           , struct  => [ "substitutionGroup $type:"
-                        , map { "   $_" } sort keys %do ]
-           , example => 'HASH with one pair'
+           , tag     => $do[1][0]
+           , struct  => [ "substitutionGroup $type:", map { "   $_" } @tags ]
+           , example => "{ $tags[0] => {...} }"
            }
         };
 }
